import { ResourceType, VariableInfo, WgslReflect } from 'wgsl_reflect'
import { WebGPUObjectError } from '../utils/errors.js'
import { BindingMap, GroupBindingMap, VariableStageInfo } from '../utils/bindings.js'

import {
  accessToBufferType,
  accessToStorageTextureAccess,
  parseTextureType,
  typeToSamplerBindingType,
  typeToTextureSampleType,
  typeToViewDimension,
  wgslToWgpuFormat
} from '../utils/wgsl-to-wgpu.js'
import { BufferBindingType } from '../enum.js'
import { BitFlags } from '../utils/bitflags.js'

/** @import { WGSLAccess, WGSLSamplerType } from '../utils/wgsl-to-wgpu.js' */

export class ShaderModule {
  _handle
  _code

  /** @type {WgslReflect | undefined} */
  _reflection

  get handle() {
    return this._handle
  }

  get label() {
    return this._handle.label
  }

  /**
   * @param {GPUDevice} device
   * @param {string} code
   * @param {string} [label]
   */
  constructor(device, code, label) {
    this._code = code

    try {
      this._handle = device.createShaderModule({ code, label })
    } catch (err) {
      throw WebGPUObjectError.from(err, ShaderModule)
    }
  }

  reflect() {
    if (this._reflection == null) {
      this._reflection = new WgslReflect(this._code)
      // no longer needed allow the GC to collect it
      this._code = undefined
    }

    return this._reflection
  }
}

export class ReflectedShader {
  _shader

  /**
   * @param {ShaderModule} shader
   */
  constructor(shader) {
    this._shader = shader
  }

  /**
   * @returns {GPUShaderStageFlags}
   */
  getShaderStages() {
    const entry = this._shader.reflect().entry
    let stages = 0

    stages |= entry.vertex.length > 0 ?
      GPUShaderStage.VERTEX : 0

    stages |= entry.fragment.length > 0 ?
      GPUShaderStage.FRAGMENT : 0

    stages |= entry.compute.length > 0 ?
      GPUShaderStage.COMPUTE : 0

    return stages
  }

  /**
   * @param {GPUShaderStageFlags} stages
   * @param {GroupBindingMap} [out=new GroupBindingMap()]
   */
  getBindingsForStage(stages, out = new GroupBindingMap()) {
    const groups = this._shader.reflect().getBindGroups()

    groups.forEach((bindings, groupIndex) => {
      if (!out.has(groupIndex)) {
        out.set(groupIndex, new BindingMap())
      }

      const bindingsMap = out.get(groupIndex)

      bindings.filter(x => x != null).forEach(variableInfo => {
        if (!bindingsMap.has(variableInfo.binding)) {
          bindingsMap.set(variableInfo.binding, { variableInfo, stages })
        } else {
          bindingsMap.get(variableInfo.binding).stages |= stages
        }
      })
    })

    return out
  }

  /**
   * @param {Map<any, any>} map 
   * @returns {number[]}
   */
  _sortKeyIndices(map) {
    return Array.from(map.keys()).sort((a, b) => a - b)
  }

  /**
   * @param {VariableInfo} _variableInfo
   * @returns {GPUBufferBindingLayout}
   */
  _parseUniform(_variableInfo) {
    return {
      type: BufferBindingType.Uniform,
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  /**
   * @param {VariableInfo} variableInfo
   * @returns {GPUBufferBindingLayout}
   */
  _parseStorage(variableInfo) {
    return {
      type: accessToBufferType(
        /** @type {WGSLAccess} */
        (variableInfo.access)
      ),
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  /**
   * @param {VariableInfo} variableInfo
   * @returns {GPUTextureBindingLayout}
   */
  _parseTexture(variableInfo) {
    const [type, sampledType] = parseTextureType(
      variableInfo.type.name
    )

    return {
      sampleType: typeToTextureSampleType(type, sampledType),
      viewDimension: typeToViewDimension(type),
      multisampled: type.includes('multisampled')
    }
  }

  /**
   * @param {VariableInfo} variableInfo 
   * @returns {GPUSamplerBindingLayout}
   */
  _parseSampler(variableInfo) {
    return {
      type: typeToSamplerBindingType(
        /** @type {WGSLSamplerType} */(variableInfo.type.name)
      )
    }
  }

  /**
   * @param {VariableInfo} variableInfo 
   * @returns {GPUStorageTextureBindingLayout}
   */
  _parseStorageTexture(variableInfo) {
    const [type] = parseTextureType(variableInfo.type.name)

    return {
      access: accessToStorageTextureAccess(
        /** @type {WGSLAccess} */(variableInfo.access)
      ),
      format: wgslToWgpuFormat(variableInfo.type.name),
      viewDimension: typeToViewDimension(type)
    }
  }

  /**
   * @param {VariableStageInfo} variableStageInfo
   * @returns {GPUBindGroupLayoutEntry}
   */
  _variableInfoToEntry(variableStageInfo) {
    const { stages: visibility, variableInfo } = variableStageInfo

    switch (variableInfo.resourceType) {
      case ResourceType.Uniform:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseUniform(variableInfo)
        }
      case ResourceType.Storage:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseStorage(variableInfo)
        }
      case ResourceType.Texture:
        return {
          binding: variableInfo.binding,
          visibility,
          texture: this._parseTexture(variableInfo)
        }
      case ResourceType.Sampler:
        return {
          binding: variableInfo.binding,
          visibility,
          sampler: this._parseSampler(variableInfo)
        }
      case ResourceType.StorageTexture:
        return {
          binding: variableInfo.binding,
          visibility,
          storageTexture: this._parseStorageTexture(variableInfo)
        }
      default:
        console.warn(`Unsupported resource type for reflection: ${ResourceType[variableInfo.resourceType]}`)
        return
    }
  }

  /**
   * @param {GroupBindingMap} groupBindings
   */
  createBindGroupLayoutEntries(groupBindings) {
    const sortedGroupIndices = this._sortKeyIndices(groupBindings)

    return sortedGroupIndices.map(groupIndex => {
      const bindingsMap = groupBindings.get(groupIndex)
      return this._sortKeyIndices(bindingsMap).map(bindingIndex => (
        this._variableInfoToEntry(bindingsMap.get(bindingIndex))
      ))
    })
      .filter(desc => desc.length > 0)
  }
}

export class UnifiedShader {
  _shader

  /**
   * @param {ReflectedShader} shader
   */
  constructor(shader) {
    this._shader = shader
  }

  createBindGroupLayoutEntries() {
    const stages = this._shader.getShaderStages()

    const unifiedShader = GPUShaderStage.VERTEX | GPUShaderStage.FRAGMENT

    if (!BitFlags.has(stages, unifiedShader)) {
      throw new Error('cant do it')
    }

    return this._shader.createBindGroupLayoutEntries(
      this._shader.getBindingsForStage(
        unifiedShader
      )
    )
  }
}

export class ReflectedShaderPair {
  _vertex
  _fragment

  /**
   * @param {ReflectedShader} vertex
   * @param {ReflectedShader} [fragment]
   */
  constructor(vertex, fragment) {
    this._vertex = vertex
    this._fragment = fragment
  }

  _createGroupBindings() {
    const groupBindings = new GroupBindingMap()

    if (
      !BitFlags.has(
        this._vertex.getShaderStages(),
        GPUShaderStage.VERTEX)
      && !BitFlags.has(
        this._fragment.getShaderStages(),
        GPUShaderStage.FRAGMENT
      )
    ) {
      throw new Error('nope')
    }

    this._vertex.getBindingsForStage(
      GPUShaderStage.VERTEX,
      groupBindings
    )

    this._fragment.getBindingsForStage(
      GPUShaderStage.FRAGMENT,
      groupBindings
    )

    return groupBindings
  }

  createBindGroupLayoutEntries() {
    // FIXME: move this call and all the other calls
    // somewhere else
    return this._shader.createBindGroupLayoutEntries(
      this._createGroupBindings()
    )
  }
}

